/* **********************************************************************************************************************
   Included Header Files
   **********************************************************************************************************************/

#include "VDEngine.h"



/* **********************************************************************************************************************
   Constructor
   **********************************************************************************************************************/

VDEngine::VDEngine() {
    mapMakerRunning = TRUE;
    npcs.clear();
};



/* **********************************************************************************************************************
   Destructor
   **********************************************************************************************************************/

VDEngine::~VDEngine() {
};



/* **********************************************************************************************************************
   Global Variables
   **********************************************************************************************************************/

// these are used to keep track of whether the windows are open or not
HWND vdEngineWindow = NULL;
HWND mapEditorWindow = NULL;
HWND npcEditorWindow = NULL;



/* **********************************************************************************************************************
   WindowProcedure
   ---------------
   Processes messages for the main window.
   **********************************************************************************************************************/

LRESULT CALLBACK WindowProcedure(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam) {
    switch (Msg) {
        case WM_CLOSE:
            // user clicked "X"
            if (hWnd == mapEditorWindow) {
                mapEditorWindow = NULL;
            } else if (hWnd == npcEditorWindow) {
                npcEditorWindow = NULL;
            };
            DestroyWindow(hWnd);
            break;
        case WM_DESTROY:
            // sent by DestroyWindow, only exit if it's the main window clicking "X"
            if (hWnd == vdEngineWindow) {
                PostQuitMessage(0);
            };
            break;
        case WM_COMMAND:
            switch (LOWORD(wParam)) {
                // Menu -> File -> Exit
                case VD_MENU_FILE_EXIT:
                    PostQuitMessage(0);
                    break;
                // Menu -> Help -> About
                case VD_MENU_HELP_ABOUT:
                    (void) MessageBox(NULL, "Varkalas D`Lonovan's 2D Game Engine", "VDEngine", MB_OK);
                    break;
                default:
                    break;
            };
            break;
        default:
            break;
    };
    return DefWindowProc(hWnd, Msg, wParam, lParam);
};



/* **********************************************************************************************************************
   Run
   ---
   Main function of the map maker.
   #param  instance  The instance of the program.
   **********************************************************************************************************************/

void VDEngine::Run(HINSTANCE instance) {
    VDMapEditor mapEditor;
    VDNPCEditor npcEditor;
    VDDirect3D graphics;
    VDInput input;
    VDWorld world;
    VDEntity user;
    VDEntity worldOffset;
    MSG message;
    int width, height;
    int choice;
    POINT mousePosition;
    RECT windowDimensions;
    RECT renderDimensions;
    char mapEditorPreviousNPCSelection[FILE_NAME_MAX_SIZE];
    char npcEditorPreviousNPCSelection[FILE_NAME_MAX_SIZE];
    char npcEditorPreviousDialogueSelection[FILE_NAME_MAX_SIZE];
    char npcEditorPreviousOptionSelection[FILE_NAME_MAX_SIZE];

    // initialize variables
    width = 896;
    height = 672;
    user.SetCurrentLocation(0, 0);
    worldOffset.SetCurrentLocation(0, 0);
    choice = 0;
    (void) strcpy_s(mapEditorPreviousNPCSelection, "");
    (void) strcpy_s(npcEditorPreviousNPCSelection, "");
    (void) strcpy_s(npcEditorPreviousDialogueSelection, "");
    (void) strcpy_s(npcEditorPreviousOptionSelection, "");

    // register all of the windows
    RegisterWindows(instance);

    // create the application window
    vdEngineWindow = CreateVDEngineWindow(instance, VD_ENGINE_CLASS_NAME, width, height);
    if (vdEngineWindow == NULL) {
        VDError("Failed to create the window.", __FILE__, __LINE__);
    };

    // initialize the graphics API
    graphics.InitializeDirect3D(vdEngineWindow);

    // initialize input
    input.InitializeInput(instance);
    input.InitializeKeyboard(vdEngineWindow);
    input.InitializeMouse(vdEngineWindow);

    // main loop
    while (mapMakerRunning) {
        // dispatch incoming sent messages, check the thread message queue, retrieve message
        while (PeekMessage(&message, NULL, 0, 0, PM_REMOVE)) {
            if (message.message == WM_QUIT) {
                return;
            } else if (message.message == WM_COMMAND) {
                // menu messages
                switch (LOWORD(message.wParam)) {
                    // Menu -> File -> Map Editor
                    case VD_MENU_FILE_MAP_EDITOR:
                        if (mapEditorWindow == NULL) {
                            // load up the NPC's
                            LoadNPCs();

                            // load the Map Editor interface
                            (void) strcpy_s(mapEditorPreviousNPCSelection, "");
                            mapEditorWindow = mapEditor.CreateMapEditor(instance, vdEngineWindow, VD_MAP_EDITOR_CLASS_NAME);

                            // populate the NPC list
                            for (list<VDNPC>::iterator anNPC = npcs.begin(); anNPC != npcs.end(); ++anNPC) {
                                VDNPC currentNPC = * anNPC;
                                (void) SendMessage(mapEditor.GetNPCListBox(), LB_ADDSTRING, 0, (LPARAM) currentNPC.GetName());
                            };
                        };
                        break;
                    // Menu -> File -> NPC Editor
                    case VD_MENU_FILE_NPC_EDITOR:
                        if (npcEditorWindow == NULL) {
                            // load up the NPC's
                            LoadNPCs();

                            // load the NPC Editor interface
                            (void) strcpy_s(npcEditorPreviousNPCSelection, "");
                            (void) strcpy_s(npcEditorPreviousDialogueSelection, "");
                            (void) strcpy_s(npcEditorPreviousOptionSelection, "");
                            npcEditorWindow = npcEditor.CreateNPCEditor(instance, vdEngineWindow, VD_NPC_EDITOR_CLASS_NAME);

                            // populate the NPC list
                            for (list<VDNPC>::iterator anNPC = npcs.begin(); anNPC != npcs.end(); ++anNPC) {
                                VDNPC currentNPC = * anNPC;
                                (void) SendMessage(npcEditor.GetNPCListBox(), LB_ADDSTRING, 0, (LPARAM) currentNPC.GetName());
                            };
                        };
                        break;
                    default:
                        break;
                };
            };
            TranslateMessage(&message);
            DispatchMessage(&message);
        };

        // gather input
        input.RetrieveLostInput();
        input.UpdateInput();
        GetCursorPos(&mousePosition);

        // take a look at the current window location
        (void) GetWindowRect(vdEngineWindow, &windowDimensions);

        // acquire the real mouse position
        int actualMousePositionX = mousePosition.x - GetSystemMetrics(SM_CXBORDER) - windowDimensions.left;
        int actualMousePositionY = mousePosition.y - (GetSystemMetrics(SM_CYBORDER) + GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CYMENU)) - windowDimensions.top;



        //////////////////////////
        // Map Editor interface //
        //////////////////////////

        if (mapEditorWindow != NULL) {

            // Save Map button
            //////////////////

            if (SendMessage(mapEditor.GetSaveMapButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
                char text[FILE_NAME_MAX_SIZE];

                // make sure there's a name in the edit box field
                (void) GetWindowText(mapEditor.GetMapNameEditBox(), text, sizeof(text));
                if (strcmp(text, "") != 0) {
                    // append .vd to the name and save the world
                    (void) strcat_s(text, ".vdm");
                    world.WriteWorld(text);
                } else {
                    MessageBox(vdEngineWindow, "No map name was provided.", "VDEngine", MB_OK);
                };

                // make sure we're not still focused on the button
                (void) SetFocus(mapEditor.GetWindow());
            };

            // Load Map button
            //////////////////

            if (SendMessage(mapEditor.GetLoadMapButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
                char text[FILE_NAME_MAX_SIZE];

                // open a world file
                VDOpenFile(mapEditor.GetWindow(), "VDEngine Maps\0*.vdm\0", text);

                // make sure the user selected a file
                if (strcmp(text, "") != 0) {
                    char actualName[FILE_NAME_MAX_SIZE];

                    // update the window
                    int n = 0, m = 0, p = 0, numSlashes;
                    while (text[n] != '\0') {
                        if (text[n] == '\\') {
                            m++;
                        };
                        n++;
                    };
                    numSlashes = m;
                    n = 0;
                    m = 0;
                    while (text[n] != '.') {
                        if (text[n] == '\\') {
                            m++;
                        };
                        n++;
                        if (text[n] != '.' && m == numSlashes) {
                            actualName[p] = text[n];
                            p++;
                        };
                    };
                    actualName[p] = '\0';
                    (void) SetWindowText(mapEditor.GetMapNameEditBox(), actualName);

                    // reset the world
                    user.SetCurrentLocation(0, 0);
                    worldOffset.SetCurrentLocation(0, 0);
                    world.ClearWorld();

                    // load the world
                    world.LoadWorld(text);

                    // clear the graphics
                    graphics.ClearScreen();

                    // add the tile sprites
                    for (int x = 0; x < world.GetWorldSizeX(); x++) {
                        for (int y = 0; y < world.GetWorldSizeY(); y++) {
                            graphics.AddSprite(world.GetTileImageFileLocation(x, y),
                                               (float) x * 32.0f,
                                               (float) y * 32.0f,
                                               32,
                                               32,
                                               FALSE,
                                               FALSE);
                        };
                    };

                    // add the NPC sprites
                    for (unsigned int n = 0; n < world.GetNPCsSize(); n++) {
                        graphics.AddSprite(world.GetNPCImageFileLocation(n),
                                          (float) world.GetNPCCurrentLocationX(n) * 32.0f,
                                          (float) world.GetNPCCurrentLocationY(n) * 32.0f,
                                          32,
                                          32,
                                          FALSE,
                                          TRUE);
                    };
                };

                // make sure we're not still focused on the button
                (void) SetFocus(mapEditor.GetWindow());
            };

            // Clear Map button
            ///////////////////

            if (SendMessage(mapEditor.GetClearMapButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
                // update the window
                (void) SetWindowText(mapEditor.GetMapNameEditBox(), "");
                (void) SetWindowText(mapEditor.GetNewAreaStatic(), "");
                (void) SetWindowText(mapEditor.GetNewAreaStartingPointX(), "0");
                (void) SetWindowText(mapEditor.GetNewAreaStartingPointY(), "0");
                mapEditor.SetNPCSprite(NULL);
                mapEditor.SetTileSprite(NULL);
                mapEditor.SetNewAreaLocation("");
                mapEditor.SetNPCSpriteLocation("");
                mapEditor.SetTileSpriteLocation("");

                // reset the world
                user.SetCurrentLocation(0, 0);
                worldOffset.SetCurrentLocation(0, 0);
                world.ClearWorld();

                // clear the graphics
                graphics.ClearScreen();

                // make sure we're not still focused on the button
                (void) SetFocus(mapEditor.GetWindow());
            };

            // Load Sprite button
            /////////////////////

            if (SendMessage(mapEditor.GetLoadSpriteButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
                char text[FILE_NAME_MAX_SIZE];
                RECT drawingArea;

                // open a sprite image
                VDOpenFile(mapEditor.GetWindow(), "Sprites\0*.PNG\0All Files\0*.*\0", text);

                // make sure the user selected a file
                if (strcmp(text, "") != 0) {
                    // set the sprite file location
                    mapEditor.SetTileSpriteLocation(text);

                    // get the sprite's dimensions to pass into VDSprite()
                    (void) mapEditor.GetTileSpriteDimensions(&drawingArea);

                    // add the sprite to the editor so it may be drawn
                    mapEditor.SetTileSprite(new VDSprite(graphics.GetDirect3DRenderer(), text, 32, 32, D3DXVECTOR3(0.0f, 0.0f, 0.0f), D3DXVECTOR3((float) drawingArea.left, (float) drawingArea.top, 0.0f), D3DCOLOR_ARGB(255, 255, 255, 255), FALSE, FALSE));
                };

                // make sure we're not still focused on the button
                (void) SetFocus(mapEditor.GetWindow());
            };

            // Load New Area button
            ///////////////////////

            if (SendMessage(mapEditor.GetLoadNewAreaButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
                char text[FILE_NAME_MAX_SIZE];

                // open the new world file
                VDOpenFile(mapEditor.GetWindow(), "VDEngine Maps\0*.vdm\0", text);

                // make sure the user selected a file
                if (strcmp(text, "") != 0) {
                    char actualName[FILE_NAME_MAX_SIZE];

                    // set the new area location
                    mapEditor.SetNewAreaLocation(text);

                    // update the window
                    int n = 0, m = 0, p = 0, numSlashes;
                    while (text[n] != '\0') {
                        if (text[n] == '\\') {
                            m++;
                        };
                        n++;
                    };
                    numSlashes = m;
                    n = 0;
                    m = 0;
                    while (text[n] != '\0') {
                        if (text[n] == '\\') {
                            m++;
                        };
                        n++;
                        if (text[n] != '\0' && m == numSlashes) {
                            actualName[p] = text[n];
                            p++;
                        };
                    };
                    actualName[p] = '\0';
                    SetWindowText(mapEditor.GetNewAreaStatic(), actualName);
                };

                // make sure we're not still focused on the button
                (void) SetFocus(mapEditor.GetWindow());
            };

            // user has an NPC selected
            ///////////////////////////

            int currentNPCIndex = (int) SendMessage(mapEditor.GetNPCListBox(), LB_GETCURSEL, 0, 0);
            if (currentNPCIndex != LB_ERR) {
                char currentNPCSelection[FILE_NAME_MAX_SIZE];

                // get the name of the current selected NPC
                (void) strcpy_s(currentNPCSelection, "");
                (void) SendMessage(mapEditor.GetNPCListBox(), LB_GETTEXT, (WPARAM) currentNPCIndex, (LPARAM) &currentNPCSelection);

                // user selects a different NPC
                ///////////////////////////////

                if (strcmp(mapEditorPreviousNPCSelection, currentNPCSelection) != 0) {
                    RECT drawingArea;

                    // copy the current selection to previous
                    (void) strcpy_s(mapEditorPreviousNPCSelection, currentNPCSelection);

                    // make sure the editor knows the file location
                    mapEditor.SetNPCSpriteLocation(GetNPC(currentNPCSelection).GetImageFileLocation());

                    // update the window
                    (void) mapEditor.GetNPCSpriteDimensions(&drawingArea);
                    mapEditor.SetNPCSprite(new VDSprite(graphics.GetDirect3DRenderer(), mapEditor.GetNPCSpriteLocation(), 32, 32, D3DXVECTOR3(0.0f, 0.0f, 0.0f), D3DXVECTOR3((float) drawingArea.left, (float) drawingArea.top, 0.0f), D3DCOLOR_ARGB(255, 255, 255, 255), FALSE, FALSE));
                };
            };

            // world editing
            ////////////////

            // cursor is on the map area
            if (actualMousePositionX >= 0 && actualMousePositionY < width) {
                // user clicks on the map
                if (input.GetMouseInput().rgbButtons[0] & 0x80) {
                    int optionStatus = -1;

                    // see if we're adding a tile or an NPC
                    if (SendMessage(mapEditor.GetNullTileOption(), BM_GETCHECK, 0, 0) == BST_CHECKED) {
                        optionStatus = 0;
                    } else if (SendMessage(mapEditor.GetOpenTileOption(), BM_GETCHECK, 0, 0) == BST_CHECKED) {
                        optionStatus = 1;
                    } else if (SendMessage(mapEditor.GetImpassableTileOption(), BM_GETCHECK, 0, 0) == BST_CHECKED) {
                        optionStatus = 2;
                    } else if (SendMessage(mapEditor.GetNewAreaTileOption(), BM_GETCHECK, 0, 0) == BST_CHECKED) {
                        optionStatus = 3;
                    } else if (SendMessage(mapEditor.GetNullNPCOption(), BM_GETCHECK, 0, 0) == BST_CHECKED) {
                        optionStatus = 4;
                    } else if (SendMessage(mapEditor.GetNormalNPCOption(), BM_GETCHECK, 0, 0) == BST_CHECKED) {
                        optionStatus = 5;
                    } else if (SendMessage(mapEditor.GetStaticNPCOption(), BM_GETCHECK, 0, 0) == BST_CHECKED) {
                        optionStatus = 6;
                    };

                    // get the current coordinates relative to the user's position (when they move with arrow keys)
                    float onScreenX = (float) user.GetCurrentLocationX() * 32.0f + actualMousePositionX - (actualMousePositionX % 32);
                    float onScreenY = (float) user.GetCurrentLocationY() * 32.0f + actualMousePositionY - (actualMousePositionY % 32);
                    int worldX = 0, worldY = 0;

                    // set the world coordinates in relation to the screen coordinates and the world view
                    if (world.GetWorldSizeX() == 0 && world.GetWorldSizeY() == 0) {
                        // first tile of the map, set the offset of the world coordinates
                        worldOffset.SetCurrentLocation((int) onScreenX / 32, (int) onScreenY / 32);
                    } else {
                        worldX = (int) onScreenX / 32 - worldOffset.GetCurrentLocationX();
                        worldY = (int) onScreenY / 32 - worldOffset.GetCurrentLocationY();
                    };

                    // TODO: REMOVE
                    graphics.RemoveTexts();
                    char testess[100];
                    sprintf_s(testess, "World Size: %d, %d\n", world.GetWorldSizeX(), world.GetWorldSizeY());
                    (void) graphics.AddText(testess, 10, 10, 80, 18, D3DCOLOR_ARGB(255, 255, 255, 255));
                    char worldCoord[100];
                    sprintf_s(worldCoord, "World Position: %d, %d\n", worldX, worldY);
                    (void) graphics.AddText(worldCoord, 10, 30, 80, 48, D3DCOLOR_ARGB(255, 255, 255, 255));

                    // reallocate the world if we have a sprite and it's not a Null option
                    if ((mapEditor.GetTileSprite() != NULL && optionStatus >= 1 && optionStatus <= 3)
                     || (mapEditor.GetNPCSprite() != NULL && optionStatus >= 5 && optionStatus <= 6)) {
                        // see if the tile is out of the current world bounds
                        if (world.CheckNewLocation(0, worldX, worldY) == NULL) {
                            // reallocate the world
                            int method = world.ReallocateWorld(worldX, worldY);

                            // we have to make sure the world view is offset since we changed the world
                            switch (method) {
                                case 1:
                                    worldOffset.SetCurrentLocation(worldOffset.GetCurrentLocationX() + worldX,
                                                                   worldOffset.GetCurrentLocationY());
                                    break;
                                case 2:
                                    worldOffset.SetCurrentLocation(worldOffset.GetCurrentLocationX(),
                                                                   worldOffset.GetCurrentLocationY() + worldY);
                                    break;
                                case 3:
                                    worldOffset.SetCurrentLocation(worldOffset.GetCurrentLocationX() + worldX,
                                                                   worldOffset.GetCurrentLocationY() + worldY);
                                    break;
                                default:
                                    break;
                            };
                        };

                        // the tile was in the negative world space, so it's 0 in the new world space
                        if (worldX < 0) {
                            worldX = 0;
                        };
                        if (worldY < 0) {
                            worldY = 0;
                        };
                    };

                    // setting the tile to NULL
                    if (optionStatus == 0) {
                        // make sure we're not out of bounds
                        if (world.CheckNewLocation(0, worldX, worldY) != NULL) {
                            // set the tile status
                            world.SetTileStatus(worldX, worldY, optionStatus);

                            // set the name of the texture in world
                            world.SetTileImageFileLocation(worldX, worldY, "");

                            // set the name and starting location of the new area
                            world.SetTileNewArea(worldX, worldY, "", 0, 0);

                            // check if there's a tile already in that spot and remove it if there is
                            graphics.RemoveTileSprite(onScreenX, onScreenY);
                        };
                    };

                    // adding a tile
                    if (mapEditor.GetTileSprite() != NULL && optionStatus >= 1 && optionStatus <= 3) {
                        // set the tile status
                        world.SetTileStatus(worldX, worldY, optionStatus);

                        // set the name of the texture in world
                        world.SetTileImageFileLocation(worldX, worldY, mapEditor.GetTileSpriteLocation());

                        // set the name of the new area
                        if (optionStatus == 3) {
                            if (strcmp(mapEditor.GetNewAreaLocation(), "") == 0) {
                                MessageBox(vdEngineWindow, "This tile will not be set as a new area as there is no new area location provided.\nPlease click on the Load New Area button and select a map first.", "VDEngine", MB_OK);
                            } else {
                                char location[10];
                                int startingPointX = 0, startingPointY = 0;

                                // get the new area starting point
                                (void) GetWindowText(mapEditor.GetNewAreaStartingPointX(), location, sizeof(location));
                                startingPointX = atoi(location);
                                (void) GetWindowText(mapEditor.GetNewAreaStartingPointY(), location, sizeof(location));
                                startingPointY = atoi(location);

                                // set the new area for the tile
                                world.SetTileNewArea(worldX, worldY, mapEditor.GetNewAreaLocation(), startingPointX, startingPointY);
                            };
                        };

                        // check if there's a tile already in that spot and remove it if there is
                        graphics.RemoveTileSprite(onScreenX, onScreenY);

                        // add the new sprite
                        graphics.AddSprite(mapEditor.GetTileSpriteLocation(), onScreenX, onScreenY, 32, 32, FALSE, FALSE);
                    };

                    // setting the NPC to NULL
                    if (optionStatus == 4) {
                        // make sure we're not out of bounds
                        if (world.CheckNewLocation(0, worldX, worldY) != NULL) {
                            // remove the NPC from the world
                            world.RemoveNPC(worldX, worldY);

                            // check if there's an NPC already in that spot and remove it if there is
                            graphics.RemoveNPCSprite(onScreenX, onScreenY);
                        };
                    };

                    // adding an NPC
                    if (mapEditor.GetNPCSprite() != NULL && optionStatus >= 5 && optionStatus <= 6) {
                        char npcName[FILE_NAME_MAX_SIZE];

                        // get the NPC name
                        (void) SendMessage(mapEditor.GetNPCListBox(), LB_GETTEXT, (WPARAM) currentNPCIndex, (LPARAM) &npcName);

                        // add the NPC to the world
                        // TODO: option status
                        world.AddNPC(npcName, worldX, worldY, mapEditor.GetNPCSpriteLocation(), optionStatus);

                        // check if there's an NPC already in that spot and remove it if there is
                        graphics.RemoveNPCSprite(onScreenX, onScreenY);

                        // add the new sprite
                        graphics.AddSprite(mapEditor.GetNPCSpriteLocation(), onScreenX, onScreenY, 32, 32, FALSE, TRUE);
                    };
                };
            };
        };



        //////////////////////////
        // NPC Editor interface //
        //////////////////////////

        if (npcEditorWindow != NULL) {

            // Save Changes button
            //////////////////////

            if (SendMessage(npcEditor.GetSaveChangesButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
                WriteNPCs();

                // make sure we're not still focused on the button
                (void) SetFocus(npcEditor.GetWindow());
            };

            // Load Sprite button
            /////////////////////

            if (SendMessage(npcEditor.GetLoadSpriteButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
                char text[FILE_NAME_MAX_SIZE];
                RECT drawingArea;

                // open a sprite image
                VDOpenFile(npcEditor.GetWindow(), "Sprites\0*.PNG\0All Files\0*.*\0", text);

                // make sure the user selected a file
                if (strcmp(text, "") != 0) {
                    // set the sprite file location
                    npcEditor.SetNPCSpriteLocation(text);

                    // get the sprite's dimensions to pass into VDSprite()
                    (void) npcEditor.GetNPCSpriteDimensions(&drawingArea);

                    // add the sprite to the editor so that it may be drawn
                    npcEditor.SetNPCSprite(new VDSprite(graphics.GetDirect3DRenderer(), text, 32, 32, D3DXVECTOR3(0.0f, 0.0f, 0.0f), D3DXVECTOR3((float) drawingArea.left, (float) drawingArea.top, 0.0f), D3DCOLOR_ARGB(255, 255, 255, 255), FALSE, FALSE));
                };

                // make sure we're not still focused on the button
                (void) SetFocus(npcEditor.GetWindow());
            };

            // Save NPC button
            //////////////////

            if (SendMessage(npcEditor.GetSaveNPCButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
                char npcText[FILE_NAME_MAX_SIZE];

                // check for a name, ID, and sprite
                (void) GetWindowText(npcEditor.GetNameEditBox(), npcText, sizeof(npcText));
                if (strcmp(npcText, "") != 0 && npcEditor.GetNPCSprite() != NULL) {
                    // add the NPC
                    AddNPC(npcText, npcEditor.GetNPCSpriteLocation());

                    // update the list box
                    (void) SendMessage(npcEditor.GetNPCListBox(), LB_ADDSTRING, 0, (LPARAM) npcText);

                    // reset the window
                    (void) SendMessage(npcEditor.GetNPCListBox(), LB_SETCURSEL, (WPARAM) - 1, 0);
                    (void) SetWindowText(npcEditor.GetNameEditBox(), "");
                    npcEditor.SetNPCSprite(NULL);
                    (void) SendMessage(npcEditor.GetDialogueListBox(), LB_RESETCONTENT, 0, 0);
                    (void) SetWindowText(npcEditor.GetTextEditBox(), "");
                    (void) SetWindowText(npcEditor.GetTextIDEditBox(), "");
                    (void) SendMessage(npcEditor.GetOptionsListBox(), LB_RESETCONTENT, 0, 0);
                    (void) SetWindowText(npcEditor.GetTextEditBox2(), "");
                    (void) SetWindowText(npcEditor.GetTextIDEditBox2(), "");
                    (void) strcpy_s(npcEditorPreviousDialogueSelection, "");
                    (void) strcpy_s(npcEditorPreviousOptionSelection, "");
                };

                // make sure we're not still focused on the button
                (void) SetFocus(npcEditor.GetNPCListBox());
            };

            // Remove NPC button
            ////////////////////

            if (SendMessage(npcEditor.GetRemoveNPCButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
                char npcText[FILE_NAME_MAX_SIZE];

                // get the index of the selected NPC text
                LRESULT index = SendMessage(npcEditor.GetNPCListBox(), LB_GETCURSEL, 0, 0);
                if (index != LB_ERR && index >= 0) {
                    // get the name of the selected NPC
                    (void) SendMessage(npcEditor.GetNPCListBox(), LB_GETTEXT, (WPARAM) index, (LPARAM) &npcText);

                    // remove the NPC from the world
                    RemoveNPC(npcText);

                    // update the list box
                    (void) SendMessage(npcEditor.GetNPCListBox(), LB_DELETESTRING, (WPARAM) index, 0);

                    // reset the window
                    (void) SendMessage(npcEditor.GetNPCListBox(), LB_SETCURSEL, (WPARAM) - 1, 0);
                    (void) SetWindowText(npcEditor.GetNameEditBox(), "");
                    npcEditor.SetNPCSprite(NULL);
                    (void) SendMessage(npcEditor.GetDialogueListBox(), LB_RESETCONTENT, 0, 0);
                    (void) SetWindowText(npcEditor.GetTextEditBox(), "");
                    (void) SetWindowText(npcEditor.GetTextIDEditBox(), "");
                    (void) SendMessage(npcEditor.GetOptionsListBox(), LB_RESETCONTENT, 0, 0);
                    (void) SetWindowText(npcEditor.GetTextEditBox2(), "");
                    (void) SetWindowText(npcEditor.GetTextIDEditBox2(), "");
                    (void) strcpy_s(npcEditorPreviousDialogueSelection, "");
                    (void) strcpy_s(npcEditorPreviousOptionSelection, "");
                };

                // make sure we're not still focused on the button
                (void) SetFocus(npcEditor.GetWindow());
            };

            // user has an NPC selected
            ///////////////////////////

            int currentNPCIndex = (int) SendMessage(npcEditor.GetNPCListBox(), LB_GETCURSEL, 0, 0);
            if (currentNPCIndex != LB_ERR) {
                char currentNPCSelection[FILE_NAME_MAX_SIZE];

                // get the name of the current selected NPC
                (void) strcpy_s(currentNPCSelection, "");
                (void) SendMessage(npcEditor.GetNPCListBox(), LB_GETTEXT, (WPARAM) currentNPCIndex, (LPARAM) &currentNPCSelection);

                // user selects a different NPC
                ///////////////////////////////

                if (strcmp(npcEditorPreviousNPCSelection, currentNPCSelection) != 0) {
                    RECT drawingArea;

                    // copy the current selection to previous
                    (void) strcpy_s(npcEditorPreviousNPCSelection, currentNPCSelection);

                    // make sure the editor knows the file location
                    // TODO:  I still don't like the returning of an instance of a class (VDNPC)... it seems unsafe, think about fixing it up sometime
                    npcEditor.SetNPCSpriteLocation(GetNPC(currentNPCSelection).GetImageFileLocation());

                    // update the window
                    (void) SetWindowText(npcEditor.GetNameEditBox(), currentNPCSelection);
                    (void) npcEditor.GetNPCSpriteDimensions(&drawingArea);
                    npcEditor.SetNPCSprite(new VDSprite(graphics.GetDirect3DRenderer(), npcEditor.GetNPCSpriteLocation(), 32, 32, D3DXVECTOR3(0.0f, 0.0f, 0.0f), D3DXVECTOR3((float) drawingArea.left, (float) drawingArea.top, 0.0f), D3DCOLOR_ARGB(255, 255, 255, 255), FALSE, FALSE));

                    // reset the window
                    (void) SendMessage(npcEditor.GetDialogueListBox(), LB_RESETCONTENT, 0, 0);
                    (void) SetWindowText(npcEditor.GetTextEditBox(), "");
                    (void) SetWindowText(npcEditor.GetTextIDEditBox(), "");
                    (void) SendMessage(npcEditor.GetOptionsListBox(), LB_RESETCONTENT, 0, 0);
                    (void) SetWindowText(npcEditor.GetTextEditBox2(), "");
                    (void) SetWindowText(npcEditor.GetTextIDEditBox2(), "");
                    (void) strcpy_s(npcEditorPreviousDialogueSelection, "");
                    (void) strcpy_s(npcEditorPreviousOptionSelection, "");

                    // populate the dialogue list
                    if (SendMessage(npcEditor.GetDialogueListBox(), LB_GETCOUNT, 0, 0) == 0) {
                        // see how many dialogues the NPC has
                        size_t numberOfDialogues = GetNPC(currentNPCSelection).GetDialoguesSize();

                        // add the dialogues to the list
                        while (numberOfDialogues > 0) {
                            char dialogue[FILE_NAME_MAX_SIZE];

                            // set it up so it's "ID: Text"
                            (void) _itoa_s(GetNPCDialogueID(currentNPCSelection, numberOfDialogues - 1), dialogue, FILE_NAME_MAX_SIZE, 10);
                            (void) strcat_s(dialogue, ": ");
                            (void) strcat_s(dialogue, GetNPCDialogue(currentNPCSelection, numberOfDialogues - 1));
                            (void) SendMessage(npcEditor.GetDialogueListBox(), LB_ADDSTRING, 0, (LPARAM) dialogue);
                            numberOfDialogues--;
                        };
                    };
                };

                // Save Text (dialogue) button
                //////////////////////////////

                if (SendMessage(npcEditor.GetSaveDialogueButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
                    char text[FILE_NAME_MAX_SIZE], textID[FILE_NAME_MAX_SIZE];

                    // check for text and ID
                    (void) GetWindowText(npcEditor.GetTextEditBox(), text, sizeof(text));
                    (void) GetWindowText(npcEditor.GetTextIDEditBox(), textID, sizeof(textID));
                    if (strcmp(text, "") != 0 && strcmp(textID, "") != 0) {
                        // add the dialogue
                        AddNPCDialogue(currentNPCSelection, atoi(textID), text);

                        // reset the window
                        (void) SendMessage(npcEditor.GetDialogueListBox(), LB_RESETCONTENT, 0, 0);
                        (void) SetWindowText(npcEditor.GetTextEditBox(), "");
                        (void) SetWindowText(npcEditor.GetTextIDEditBox(), "");
                        (void) SendMessage(npcEditor.GetOptionsListBox(), LB_RESETCONTENT, 0, 0);
                        (void) SetWindowText(npcEditor.GetTextEditBox2(), "");
                        (void) SetWindowText(npcEditor.GetTextIDEditBox2(), "");
                        (void) strcpy_s(npcEditorPreviousDialogueSelection, "");
                        (void) strcpy_s(npcEditorPreviousOptionSelection, "");

                        // populate the dialogue list
                        if (SendMessage(npcEditor.GetDialogueListBox(), LB_GETCOUNT, 0, 0) == 0) {
                            // see how many dialogues the NPC has
                            size_t numberOfDialogues = GetNPC(currentNPCSelection).GetDialoguesSize();

                            // add the dialogues to the list
                            while (numberOfDialogues > 0) {
                                char dialogue[FILE_NAME_MAX_SIZE];

                                // set it up so it's "ID: Text"
                                (void) _itoa_s(GetNPCDialogueID(currentNPCSelection, numberOfDialogues - 1), dialogue, FILE_NAME_MAX_SIZE, 10);
                                (void) strcat_s(dialogue, ": ");
                                (void) strcat_s(dialogue, GetNPCDialogue(currentNPCSelection, numberOfDialogues - 1));
                                (void) SendMessage(npcEditor.GetDialogueListBox(), LB_ADDSTRING, 0, (LPARAM) dialogue);
                                numberOfDialogues--;
                            };
                        };
                    };

                    // make sure we're not still focused on the button
                    (void) SetFocus(npcEditor.GetWindow());
                };

                // Remove Dialogue button
                /////////////////////////

                if (SendMessage(npcEditor.GetRemoveDialogueButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
                    char text[FILE_NAME_MAX_SIZE];

                    // get the index of the selected Dialogue text
                    LRESULT index = SendMessage(npcEditor.GetDialogueListBox(), LB_GETCURSEL, 0, 0);
                    if (index != LB_ERR && index >= 0) {
                        char id[10];

                        // get the name of the selected dialogue
                        (void) SendMessage(npcEditor.GetDialogueListBox(), LB_GETTEXT, (WPARAM) index, (LPARAM) &text);

                        // get the dialogue ID, remove the dialogue from the NPC
                        int n = 0;
                        while (text[n] != ':') {
                            id[n] = text[n];
                            n++;
                        };
                        id[n] = '\0';
                        RemoveNPCDialogue(currentNPCSelection, atoi(id));

                        // update the list box
                        (void) SendMessage(npcEditor.GetDialogueListBox(), LB_DELETESTRING, (WPARAM) index, 0);
                        (void) SendMessage(npcEditor.GetDialogueListBox(), LB_SETCURSEL, (WPARAM) - 1, 0);

                        // reset the window
                        (void) SetWindowText(npcEditor.GetTextEditBox(), "");
                        (void) SetWindowText(npcEditor.GetTextIDEditBox(), "");
                        (void) SendMessage(npcEditor.GetOptionsListBox(), LB_RESETCONTENT, 0, 0);
                        (void) SetWindowText(npcEditor.GetTextEditBox2(), "");
                        (void) SetWindowText(npcEditor.GetTextIDEditBox2(), "");
                        (void) strcpy_s(npcEditorPreviousDialogueSelection, "");
                        (void) strcpy_s(npcEditorPreviousOptionSelection, "");
                    };

                    // make sure we're not still focused on the button
                    (void) SetFocus(npcEditor.GetWindow());
                };

                // user has a dialogue selected
                ///////////////////////////////

                int currentDialogueIndex = (int) SendMessage(npcEditor.GetDialogueListBox(), LB_GETCURSEL, 0, 0);
                if (currentDialogueIndex != LB_ERR) {
                    char id[10], currentDialogueSelection[FILE_NAME_MAX_SIZE], dialogueName[FILE_NAME_MAX_SIZE];

                    // get the string of the current selected dialogue
                    (void) strcpy_s(currentDialogueSelection, "");
                    (void) SendMessage(npcEditor.GetDialogueListBox(), LB_GETTEXT, (WPARAM) currentDialogueIndex, (LPARAM) &currentDialogueSelection);

                    // get the ID and name of the dialogue
                    int n = 0, m = 0;
                    while (currentDialogueSelection[n] != ':') {
                        id[n] = currentDialogueSelection[n];
                        n++;
                    };
                    id[n] = '\0';
                    n++;
                    while (currentDialogueSelection[n] != ' ') {
                        n++;
                    };
                    n++;
                    while (currentDialogueSelection[n] != '\0') {
                        dialogueName[m] = currentDialogueSelection[n];
                        n++;
                        m++;
                    };
                    dialogueName[m] = '\0';
                    int currentDialogueID = atoi(id);

                    // user selects a different dialogue
                    ////////////////////////////////////

                    if (strcmp(npcEditorPreviousDialogueSelection, currentDialogueSelection) != 0) {
                        // copy the current selection to previous
                        (void) strcpy_s(npcEditorPreviousDialogueSelection, currentDialogueSelection);

                        // update the window
                        (void) SetWindowText(npcEditor.GetTextEditBox(), dialogueName);
                        (void) SetWindowText(npcEditor.GetTextIDEditBox(), id);
                        (void) SendMessage(npcEditor.GetOptionsListBox(), LB_RESETCONTENT, 0, 0);
                        (void) SetWindowText(npcEditor.GetTextEditBox2(), "");
                        (void) SetWindowText(npcEditor.GetTextIDEditBox2(), "");
                        (void) strcpy_s(npcEditorPreviousOptionSelection, "");

                        // populate the options list
                        if (SendMessage(npcEditor.GetOptionsListBox(), LB_GETCOUNT, 0, 0) == 0) {
                            // see how many options the dialogue of the NPC has
                            size_t numberOfOptions = GetNPC(currentNPCSelection).GetOptionsSize(currentDialogueID);

                            // add the options to the list
                            while (numberOfOptions > 0) {
                                char option[FILE_NAME_MAX_SIZE];

                                // set it up so it's "ID: Text"
                                (void) _itoa_s(GetNPCDialogueOptionNextID(currentNPCSelection, currentDialogueID, numberOfOptions - 1), option, FILE_NAME_MAX_SIZE, 10);
                                (void) strcat_s(option, "[next]: ");
                                (void) strcat_s(option, GetNPCDialogueOption(currentNPCSelection, currentDialogueID, numberOfOptions - 1));
                                (void) SendMessage(npcEditor.GetOptionsListBox(), LB_ADDSTRING, 0, (LPARAM) option);
                                numberOfOptions--;
                            };
                        };
                    };

                    // Save Text (option) button
                    ////////////////////////////

                    if (SendMessage(npcEditor.GetSaveOptionButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
                        char text[FILE_NAME_MAX_SIZE], nextID[FILE_NAME_MAX_SIZE];

                        // check for text and ID
                        (void) GetWindowText(npcEditor.GetTextEditBox2(), text, sizeof(text));
                        (void) GetWindowText(npcEditor.GetTextIDEditBox2(), nextID, sizeof(nextID));
                        if (strcmp(text, "") != 0 && strcmp(nextID, "") != 0) {
                            // add the option
                            (void) AddNPCDialogueOption(currentNPCSelection, currentDialogueID, text, atoi(nextID));

                            // reset the window
                            (void) SendMessage(npcEditor.GetOptionsListBox(), LB_RESETCONTENT, 0, 0);
                            (void) SetWindowText(npcEditor.GetTextEditBox2(), "");
                            (void) SetWindowText(npcEditor.GetTextIDEditBox2(), "");
                            (void) strcpy_s(npcEditorPreviousOptionSelection, "");

                            // populate the options list
                            if (SendMessage(npcEditor.GetOptionsListBox(), LB_GETCOUNT, 0, 0) == 0) {
                                // see how many options the dialogue of the NPC has
                                size_t numberOfOptions = GetNPC(currentNPCSelection).GetOptionsSize(currentDialogueID);

                                // add the options to the list
                                while (numberOfOptions > 0) {
                                    char option[FILE_NAME_MAX_SIZE];

                                    // set it up so it's "ID: Text"
                                    (void) _itoa_s(GetNPCDialogueOptionNextID(currentNPCSelection, currentDialogueID, numberOfOptions - 1), option, FILE_NAME_MAX_SIZE, 10);
                                    (void) strcat_s(option, "[next]: ");
                                    (void) strcat_s(option, GetNPCDialogueOption(currentNPCSelection, currentDialogueID, numberOfOptions - 1));
                                    (void) SendMessage(npcEditor.GetOptionsListBox(), LB_ADDSTRING, 0, (LPARAM) option);
                                    numberOfOptions--;
                                };
                            };
                        };

                        // make sure we're not still focused on the button
                        (void) SetFocus(npcEditor.GetWindow());
                    };

                    // Remove Response button
                    /////////////////////////

                    if (SendMessage(npcEditor.GetRemoveOptionButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
                        char text[FILE_NAME_MAX_SIZE];

                        // get the index of the selected Option text
                        LRESULT index = SendMessage(npcEditor.GetOptionsListBox(), LB_GETCURSEL, 0, 0);
                        if (index != LB_ERR && index >= 0) {
                            char optionName[FILE_NAME_MAX_SIZE];

                            // get the string of the selected option
                            (void) SendMessage(npcEditor.GetOptionsListBox(), LB_GETTEXT, (WPARAM) index, (LPARAM) &text);

                            // get the option name, remove the option from the dialogue
                            int n = 0, m = 0;
                            while (text[n] != ' ') {
                                n++;
                            };
                            n++;
                            while (text[n] != '\0') {
                                optionName[m] = text[n];
                                n++;
                                m++;
                            };
                            optionName[m] = '\0';
                            RemoveNPCDialogueOption(currentNPCSelection, currentDialogueID, optionName);

                            // update the list box
                            (void) SendMessage(npcEditor.GetOptionsListBox(), LB_DELETESTRING, (WPARAM) index, 0);
                            (void) SendMessage(npcEditor.GetOptionsListBox(), LB_SETCURSEL, (WPARAM) - 1, 0);

                            // reset the window
                            (void) SetWindowText(npcEditor.GetTextEditBox2(), "");
                            (void) SetWindowText(npcEditor.GetTextIDEditBox2(), "");
                            (void) strcpy_s(npcEditorPreviousOptionSelection, "");
                        };

                        // make sure we're not still focused on the button
                        (void) SetFocus(npcEditor.GetWindow());
                    };

                    // user has an option selected
                    //////////////////////////////

                    int currentOptionIndex = (int) SendMessage(npcEditor.GetOptionsListBox(), LB_GETCURSEL, 0, 0);
                    if (currentOptionIndex != LB_ERR) {
                        char currentOptionSelection[FILE_NAME_MAX_SIZE];

                        // get the name of the current selected option
                        (void) strcpy_s(currentOptionSelection, "");
                        (void) SendMessage(npcEditor.GetOptionsListBox(), LB_GETTEXT, (WPARAM) currentOptionIndex, (LPARAM) &currentOptionSelection);

                        // user selects a different option
                        //////////////////////////////////

                        if (strcmp(npcEditorPreviousOptionSelection, currentOptionSelection) != 0) {
                            char id[10], optionName[FILE_NAME_MAX_SIZE];

                            // copy the current selection to previous
                            (void) strcpy_s(npcEditorPreviousOptionSelection, currentOptionSelection);

                            // get the option ID and name
                            int n = 0, m = 0;
                            while (currentOptionSelection[n] != '[') {
                                id[n] = currentOptionSelection[n];
                                n++;
                            };
                            id[n] = '\0';
                            while (currentOptionSelection[n] != ' ') {
                                n++;
                            };
                            n++;
                            while (currentOptionSelection[n] != '\0') {
                                optionName[m] = currentOptionSelection[n];
                                n++;
                                m++;
                            };
                            optionName[m] = '\0';

                            // update the window
                            (void) SetWindowText(npcEditor.GetTextEditBox2(), optionName);
                            (void) SetWindowText(npcEditor.GetTextIDEditBox2(), id);
                        };
                    };
                };
            };
        };



        // movement (via arrow keys)
        ////////////////////////////

        // see if any keys were pressed
        if (input.GetKeyboardInput()[DIK_ESCAPE] & 0x80) {
            mapMakerRunning = FALSE;
        } else if (input.GetKeyboardInput()[DIK_LEFT] & 0x80) {
            choice = 1;
        } else if (input.GetKeyboardInput()[DIK_RIGHT] & 0x80) {
            choice = 2;
        } else if (input.GetKeyboardInput()[DIK_UP] & 0x80) {
            choice = 3;
        } else if (input.GetKeyboardInput()[DIK_DOWN] & 0x80) {
            choice = 4;
        };

        // user decided to move
        if (choice >= 1 && choice <= 4) {
            // update the user's location
            user.Move(choice);
            choice = 0;
        };



        // drawing
        //////////

        // draw graphics
        renderDimensions.left = 0;
        renderDimensions.top = 0;
        renderDimensions.right = width - GetSystemMetrics(SM_CXBORDER);
        renderDimensions.bottom = height - GetSystemMetrics(SM_CYBORDER);
        graphics.Render(&renderDimensions, 0, 0, user.GetCurrentLocationX(), user.GetCurrentLocationY());

        // draw Map Editor sprites
        if (mapEditorWindow != NULL) {
            RECT drawingArea;
            (void) mapEditor.GetTileSpriteDimensions(&drawingArea);
            graphics.RenderSprite(&drawingArea, mapEditor.GetWindow(), mapEditor.GetTileSprite());
            (void) mapEditor.GetNPCSpriteDimensions(&drawingArea);
            graphics.RenderSprite(&drawingArea, mapEditor.GetWindow(), mapEditor.GetNPCSprite());
        };

        // draw NPC Editor sprites
        if (npcEditorWindow != NULL) {
            RECT drawingArea;
            (void) npcEditor.GetNPCSpriteDimensions(&drawingArea);
            graphics.RenderSprite(&drawingArea, npcEditor.GetWindow(), npcEditor.GetNPCSprite());
        };
    };
};



/* **********************************************************************************************************************
   RegisterWindows
   ---------------
   Registers the window classes that we'll be using in this application.
   #param  instance   Instance of the application.
   **********************************************************************************************************************/

void VDEngine::RegisterWindows(HINSTANCE instance) {
    WNDCLASSEX windowClassStructure;
    ATOM registerClassEx;

    // main window
    windowClassStructure.cbClsExtra = 0;                                   // number of extra bytes to allocate following the window-class structure
    windowClassStructure.cbSize = sizeof(WNDCLASSEX);                      // size of the structure (in bytes)
    windowClassStructure.cbWndExtra = 0;                                   // number of extra bytes to allocate following the window instance
    windowClassStructure.hbrBackground = GetSysColorBrush(COLOR_BTNFACE);  // background brush
    windowClassStructure.hCursor = LoadCursor(NULL, IDC_ARROW);            // cursor
    windowClassStructure.hIcon = NULL;                                     // icon
    windowClassStructure.hIconSm = LoadIcon(NULL, IDI_APPLICATION);        // small icon
    windowClassStructure.hInstance = instance;                             // instance that has the window procedure
    windowClassStructure.lpfnWndProc = WindowProcedure;                    // window procedure
    windowClassStructure.lpszClassName = VD_ENGINE_CLASS_NAME;             // class name
    windowClassStructure.lpszMenuName = MAKEINTRESOURCE(VD_MENU);          // menu resource name
    windowClassStructure.style = NULL;                                     // style
    registerClassEx = RegisterClassEx(&windowClassStructure);
    if (registerClassEx == 0) {
        VDError("Failed to register the window class.", __FILE__, __LINE__);
    };
    (void) SecureZeroMemory(&windowClassStructure, sizeof(windowClassStructure));

    // Map Editor
    windowClassStructure.cbClsExtra = 0;                                   // number of extra bytes to allocate following the window-class structure
    windowClassStructure.cbSize = sizeof(WNDCLASSEX);                      // size of the structure (in bytes)
    windowClassStructure.cbWndExtra = 0;                                   // number of extra bytes to allocate following the window instance
    windowClassStructure.hbrBackground = GetSysColorBrush(COLOR_BTNFACE);  // background brush
    windowClassStructure.hCursor = LoadCursor(NULL, IDC_ARROW);            // cursor
    windowClassStructure.hIcon = NULL;                                     // icon
    windowClassStructure.hIconSm = LoadIcon(NULL, IDI_APPLICATION);        // small icon
    windowClassStructure.hInstance = instance;                             // instance that has the window procedure
    windowClassStructure.lpfnWndProc = WindowProcedure;                    // window procedure
    windowClassStructure.lpszClassName = VD_MAP_EDITOR_CLASS_NAME;         // class name
    windowClassStructure.lpszMenuName = NULL;                              // menu resource name
    windowClassStructure.style = NULL;                                     // style
    registerClassEx = RegisterClassEx(&windowClassStructure);
    if (registerClassEx == 0) {
        VDError("Failed to register the window class.", __FILE__, __LINE__);
    };

    // NPC Editor
    windowClassStructure.cbClsExtra = 0;                                   // number of extra bytes to allocate following the window-class structure
    windowClassStructure.cbSize = sizeof(WNDCLASSEX);                      // size of the structure (in bytes)
    windowClassStructure.cbWndExtra = 0;                                   // number of extra bytes to allocate following the window instance
    windowClassStructure.hbrBackground = GetSysColorBrush(COLOR_BTNFACE);  // background brush
    windowClassStructure.hCursor = LoadCursor(NULL, IDC_ARROW);            // cursor
    windowClassStructure.hIcon = NULL;                                     // icon
    windowClassStructure.hIconSm = LoadIcon(NULL, IDI_APPLICATION);        // small icon
    windowClassStructure.hInstance = instance;                             // instance that has the window procedure
    windowClassStructure.lpfnWndProc = WindowProcedure;                    // window procedure
    windowClassStructure.lpszClassName = VD_NPC_EDITOR_CLASS_NAME;         // class name
    windowClassStructure.lpszMenuName = NULL;                              // menu resource name
    windowClassStructure.style = NULL;                                     // style
    registerClassEx = RegisterClassEx(&windowClassStructure);
    if (registerClassEx == 0) {
        VDError("Failed to register the window class.", __FILE__, __LINE__);
    };
};



/* **********************************************************************************************************************
   CreateVDEngineWindow
   --------------------
   Creates the main window of the engine.
   #param  instance   Instance of the application.
   #param  className  The name of the window class.
   #param  width      Width of the window.
   #param  height     Height of the window.
   #return  A handle to the window.
   **********************************************************************************************************************/

HWND VDEngine::CreateVDEngineWindow(HINSTANCE instance, LPCSTR className, int width, int height) {
    // create the menu resource
    HMENU menu = LoadMenu(instance, MAKEINTRESOURCE(VD_MENU));
    if (menu == NULL) {
        VDError("Failed to create the menu resource.", __FILE__, __LINE__);
    };

    return CreateWindowEx(NULL,                                                   // extended window styles
                          className,                                              // class name
                          "VDEngine",                                             // window name
                          WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX | WS_VISIBLE,  // window styles
                          0,                                                      // x coordinate (0 = left)
                          0,                                                      // y coordinate (0 = top)
                          width,                                                  // width
                          height,                                                 // height
                          NULL,                                                   // parent/owner window
                          menu,                                                   // menu
                          instance,                                               // instance
                          NULL);                                                  // used for CREATESTRUCT structure
};



/* **********************************************************************************************************************
   VDOpenFile
   ----------
   Creates an open file box and sets the file name parameter to the chosen file.
   #param  window    Handle to the window opening the box.
   #param  filter    The type of filter to distinguish which types of files to open.
   #param  fileName  (out) The chosen file location.
   **********************************************************************************************************************/

void VDEngine::VDOpenFile(HWND window, LPSTR filter, LPSTR fileName) {
    char file[FILE_NAME_MAX_SIZE];
    OPENFILENAME openFileBox;

    // zero out the open file box just to be on the safe side
    SecureZeroMemory(&openFileBox, sizeof(OPENFILENAME));

    // set up the open file structure
                openFileBox.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_NOCHANGEDIR;
            openFileBox.hwndOwner = window;
            openFileBox.lpstrFile = file;
         openFileBox.lpstrFile[0] = '\0';
       openFileBox.lpstrFileTitle = NULL;
          openFileBox.lpstrFilter = filter;
      openFileBox.lpstrInitialDir = NULL;
          openFileBox.lStructSize = sizeof(OPENFILENAME);
         openFileBox.nFilterIndex = 1;
             openFileBox.nMaxFile = sizeof(file);
        openFileBox.nMaxFileTitle = NULL;

    // open a file
    BOOL returnValue = GetOpenFileName(&openFileBox);
    if (returnValue != 0) {
        (void) strcpy_s(fileName, sizeof(file), file);
    } else {
        (void) strcpy_s(fileName, sizeof(file), "");
    };
};



/* **********************************************************************************************************************
   AddNPC
   ------
   Adds an NPC to the list.
   #param  name               Name of the NPC.
   #param  imageFileLocation  The file location of the sprite.
   **********************************************************************************************************************/

void VDEngine::AddNPC(LPCSTR name, LPCSTR imageFileLocation) {
    // see if we're trying to add an NPC that already exists
    for (list<VDNPC>::iterator anNPC = npcs.begin(); anNPC != npcs.end(); ++anNPC) {
        if (strcmp(name, anNPC->GetName()) == 0) {
            // only change the image, not the name
            anNPC->SetImageFileLocation(imageFileLocation);
            return;
        };
    };

    // add the NPC
    VDNPC newNPC = VDNPC(name);
    newNPC.SetImageFileLocation(imageFileLocation);
    npcs.push_back(newNPC);
};



/* **********************************************************************************************************************
   RemoveNPC
   ---------
   Removes an NPC from the list given their name.
   #param  name  Name of the NPC.
   **********************************************************************************************************************/

void VDEngine::RemoveNPC(LPCSTR name) {
    for (list<VDNPC>::iterator anNPC = npcs.begin(); anNPC != npcs.end();) {
        if (strcmp(name, anNPC->GetName()) == 0) {
            anNPC = npcs.erase(anNPC);
        } else {
            ++anNPC;
        };
    };
};



/* **********************************************************************************************************************
   AddNPCDialogue
   --------------
   Tells the NPC to add a new dialogue.
   #param  name  Name of the NPC.
   #param  id    The ID of the dialogue.
   #param  text  The text of the dialogue.
   **********************************************************************************************************************/

void VDEngine::AddNPCDialogue(LPCSTR name, int id, LPCSTR text) {
    for (list<VDNPC>::iterator anNPC = npcs.begin(); anNPC != npcs.end(); ++anNPC) {
        if (strcmp(name, anNPC->GetName()) == 0) {
            anNPC->AddDialogue(id, text);
        };
    };
};



/* **********************************************************************************************************************
   RemoveNPCDialogue
   -----------------
   Tells the NPC to remove a dialogue.
   #param  name  Name of the NPC.
   #param  id    The ID of the dialogue.
   #param  text  The text of the dialogue.
   **********************************************************************************************************************/

void VDEngine::RemoveNPCDialogue(LPCSTR name, int id) {
    for (list<VDNPC>::iterator anNPC = npcs.begin(); anNPC != npcs.end(); ++anNPC) {
        if (strcmp(name, anNPC->GetName()) == 0) {
            anNPC->RemoveDialogue(id);
        };
    };
};



/* **********************************************************************************************************************
   AddNPCDialogueOption
   --------------------
   Tells the NPC to tell the dialogue to add a new option.
   #param  npcName     Name of the NPC.
   #param  dialogueID  The ID of the dialogue.
   #param  text        The string of the option text.
   #param  nextID      ID of the next dialogue.
   **********************************************************************************************************************/

void VDEngine::AddNPCDialogueOption(LPCSTR npcName, int dialogueID, LPCSTR text, int nextID) {
    for (list<VDNPC>::iterator anNPC = npcs.begin(); anNPC != npcs.end(); ++anNPC) {
        if (strcmp(npcName, anNPC->GetName()) == 0) {
            anNPC->AddOption(dialogueID, text, nextID);
        };
    };
};



/* **********************************************************************************************************************
   RemoveNPCDialogueOption
   -----------------------
   Tells the NPC to tell the dialogue to remove an option.
   #param  npcName     Name of the NPC.
   #param  dialogueID  The ID of the dialogue.
   #param  optionName  The string of the option text.
   #param  nextID      ID of the next dialogue.
   **********************************************************************************************************************/

void VDEngine::RemoveNPCDialogueOption(LPCSTR npcName, int dialogueID, LPCSTR optionName) {
    for (list<VDNPC>::iterator anNPC = npcs.begin(); anNPC != npcs.end(); ++anNPC) {
        if (strcmp(npcName, anNPC->GetName()) == 0) {
            anNPC->RemoveOption(dialogueID, optionName);
        };
    };
};



/* **********************************************************************************************************************
   WriteNPCs
   ---------
   Saves the NPC's to hard disk.
   **********************************************************************************************************************/

void VDEngine::WriteNPCs() {
    HANDLE file = NULL;
    BOOL returnValue = FALSE;
    DWORD numberOfBytesWritten = 0;
    size_t numberOfNPCs = 0;

    // open the file
    file = CreateFile("NPCs.vd", GENERIC_WRITE, FILE_SHARE_WRITE, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    if (file == INVALID_HANDLE_VALUE) {
        VDError("Failed to create the NPCs.vd file.", __FILE__, __LINE__);
    };

    // make sure we let it know how many NPC elements we're adding
    numberOfNPCs = npcs.size();
    returnValue = WriteFile(file, &numberOfNPCs, sizeof(numberOfNPCs), &numberOfBytesWritten, NULL);
    if (returnValue == FALSE) {
        VDError("Failed to write NPC data to file.", __FILE__, __LINE__);
    };

    // write all the NPCs
    for (list<VDNPC>::iterator anNPC = npcs.begin(); anNPC != npcs.end(); ++anNPC) {
        anNPC->WriteNPC(file);
    };

    // close the file
    returnValue = CloseHandle(file);
    if (returnValue == FALSE) {
        VDError("Failed to close the NPCs.vd file.", __FILE__, __LINE__);
    };
};



/* **********************************************************************************************************************
   LoadNPCs
   --------
   Loads the NPC's into the game.
   **********************************************************************************************************************/

void VDEngine::LoadNPCs() {
    HANDLE file = NULL;
    BOOL returnValue = FALSE;
    DWORD numberOfBytesRead = 0;
    size_t numberOfNPCs = 0;

    // clear the NPC data first
    npcs.clear();

    // open the file
    file = CreateFile("NPCs.vd", GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (file == INVALID_HANDLE_VALUE) {
        return;
    };

    // get how many NPC's we're adding
    returnValue = ReadFile(file, &numberOfNPCs, sizeof(numberOfNPCs), &numberOfBytesRead, NULL);
    if (returnValue == FALSE) {
        VDError("Failed to read from the NPCs.vd file.", __FILE__, __LINE__);
    };

    // read all the NPC's
    for (unsigned int n = 0; n < numberOfNPCs; n++) {
        VDNPC newNPC;
        newNPC.LoadNPC(file);
        npcs.push_back(newNPC);
    };

    // close the file
    returnValue = CloseHandle(file);
    if (returnValue == FALSE) {
        VDError("Failed to close the NPCs.vd file.", __FILE__, __LINE__);
    };
};



/* **********************************************************************************************************************
   Accessors and Modifiers
   **********************************************************************************************************************/

VDNPC VDEngine::GetNPC(LPCSTR name) {
    for (list<VDNPC>::iterator anNPC = npcs.begin(); anNPC != npcs.end(); ++anNPC) {
        VDNPC npc = *anNPC;
        if (strcmp(name, anNPC->GetName()) == 0) {
            return *anNPC;
        };
    };

    return NULL;
};

LPCSTR VDEngine::GetNPCDialogue(LPCSTR npcName, int index) {
    for (list<VDNPC>::iterator anNPC = npcs.begin(); anNPC != npcs.end(); ++anNPC) {
        if (strcmp(npcName, anNPC->GetName()) == 0) {
            return anNPC->GetDialogueText(index);
        };
    };

    return NULL;
};

// used for getting the ID at a specified index
unsigned int VDEngine::GetNPCDialogueID(LPCSTR npcName, int index) {
    for (list<VDNPC>::iterator anNPC = npcs.begin(); anNPC != npcs.end(); ++anNPC) {
        if (strcmp(npcName, anNPC->GetName()) == 0) {
            return anNPC->GetDialogueID(index);
        };
    };

    return NULL;
};

LPCSTR VDEngine::GetNPCDialogueOption(LPCSTR npcName, int dialogueID, int optionIndex) {
    for (list<VDNPC>::iterator anNPC = npcs.begin(); anNPC != npcs.end(); ++anNPC) {
        if (strcmp(npcName, anNPC->GetName()) == 0) {
            return anNPC->GetOption(dialogueID, optionIndex);
        };
    };

    return NULL;
};

unsigned int VDEngine::GetNPCDialogueOptionNextID(LPCSTR npcName, int dialogueID, int optionIndex) {
    for (list<VDNPC>::iterator anNPC = npcs.begin(); anNPC != npcs.end(); ++anNPC) {
        if (strcmp(npcName, anNPC->GetName()) == 0) {
            return anNPC->GetOptionNextID(dialogueID, optionIndex);
        };
    };

    return NULL;
};